// Written by Tilman Neumann (www.tilman-neumann.de) 2016, 2017
// Any commercial use of this software requires the written consent of the author.
// Usage and republication are free for non-commercial purposes, as long as this header is carried along.
// The author takes no warranties for the correctness of the software.
package de.tilman_neumann.math.factor.squfof;

import java.math.BigInteger;

import org.apache.log4j.Logger;

import de.tilman_neumann.math.base.bigint.SqrtInt;
import de.tilman_neumann.math.base.bigint.sequence.IntegerSequence;
import de.tilman_neumann.math.base.bigint.sequence.SquarefreeSequence;
import de.tilman_neumann.math.factor.FactorAlgorithmBase;

import static de.tilman_neumann.math.base.bigint.BigIntConstants.*;

/**
 * Shanks' SQUFOF algorithm, 63-bit version.<br/>
 * Implemented according to <link>http://en.wikipedia.org/wiki/Shanks'_square_forms_factorization</link>.
 * 
 * Stopping criterion: after a maximum number of iterations.
 * 
 * @author Tilman Neumann
 */
public class SquFoF63_01 extends FactorAlgorithmBase {
	private static final Logger LOG = Logger.getLogger(SquFoF63_01.class);

	// input
	private BigInteger multiplier;
	private int stopRoot;
	private float stopMult;
	private BigInteger N, kN;
	private long floor_sqrt_kN;
	
	// maximum number of iterations
	private int maxI;
	
	/**
	 * Standard constructor.
	 * @param multiplier
	 * @param stopRoot order of the root to compute the maximum number of iterations
	 * @param stopMult multiplier to compute the maximum number of iterations
	 */
	public SquFoF63_01(int multiplier, int stopRoot, float stopMult) {
		this.multiplier = BigInteger.valueOf(multiplier);
		this.stopRoot = stopRoot;
		this.stopMult = stopMult;
	}

	@Override
	public String getName() {
		return "SquFoF63_01(" + multiplier + ", " + stopRoot + ", " + stopMult + ")";
	}
	
	/**
	 * Find a factor of the given composite N.</br></br>
	 * <em>Warning:</em> This method will not return when called with a prime argument.
	 * @param N composite integer
	 * @return factor of N
	 */
	public BigInteger findSingleFactor(BigInteger N) {
		this.N = N;
		// max iterations: there is no need to account for k, because expansions of smooth kN are typically not longer than those for N.
		this.maxI = (int) (stopMult*Math.pow(N.doubleValue(), 1.0/stopRoot));
		// sequence: for each new N start again with the first k
		IntegerSequence<BigInteger> kSequence = new SquarefreeSequence(multiplier);
		kSequence.reset(N);
		int iterations=0;
		while (true) {
			// get a new k, return immediately if kN is square
			this.kN = kSequence.next().multiply(N);
			BigInteger[] iSqrt = SqrtInt.iSqrt(kN);
			if (iSqrt[0].equals(iSqrt[1])) return N.gcd(iSqrt[0]);
			this.floor_sqrt_kN = iSqrt[0].longValue();
			// search square Q_i
			BigInteger factor = test();
			if (factor!=null) {
				if (iterations>1000) LOG.debug(getName() + ": Found factor after " + iterations + " k-values");
				return factor;
			}
			iterations++;
		}
	}

	protected BigInteger test() {
		// initialization for first iteration step
		int i = 0;
		long P_im1 = 1;
		long P_i = floor_sqrt_kN;
		BigInteger P_i_big = BigInteger.valueOf(P_i);
		long Q_i = 1;
		long Q_ip1 = kN.subtract(P_i_big.multiply(P_i_big)).longValue();
		
		// first iteration step
		while (true) {
			// [McMath 2004] points out that we have to look for a square Q_i at some even i.
			// Here I test Q_i+1, so I have to look for square Q_i+1 at odd i!
			if (i%2==1) {
				long Q_ip1_sqrt = (long) Math.sqrt(Q_ip1);
				if (Q_ip1_sqrt*Q_ip1_sqrt==Q_ip1) {
					// Q_i+1 is square -> do reverse iteration
					BigInteger factor = reverseIteration(P_i, Q_ip1_sqrt);
					if (factor!=null) return factor; // if factor is null we try another k
				}
			}
		
			// exit ?
			if (++i==maxI) return null;
			// keep values from last round
			P_im1 = P_i;
			long Q_im1 = Q_i;
			Q_i = Q_ip1;
			// compute next values
			long b_i = (floor_sqrt_kN + P_im1)/Q_i; // floor(rational result)
			P_i = b_i*Q_i - P_im1;
			Q_ip1 = Q_im1 + b_i*(P_im1-P_i);
		}
	}
	
	private BigInteger reverseIteration(long found_P, long found_Q_sqrt) {
		// initialization for second iteration step
		int i = 0;
		long b_i = (floor_sqrt_kN-found_P)/found_Q_sqrt; // floor(rational result)
		long P_i = b_i*found_Q_sqrt + found_P;
		long Q_i = found_Q_sqrt;
		// computing Q_ip1 needs BigInteger precision at every detail!
		BigInteger P_i_big = BigInteger.valueOf(P_i);
		long Q_ip1 = kN.subtract(P_i_big.multiply(P_i_big)).divide(BigInteger.valueOf(found_Q_sqrt)).longValue();
		
		// second iteration step
		long P_im1, Q_im1;
		do {
			// exit ?
			if (++i==maxI) return null;
			// keep values from last round
			P_im1 = P_i;
			Q_im1 = Q_i;
			Q_i = Q_ip1;
			// compute next values
			b_i = (floor_sqrt_kN+P_im1)/Q_i; // floor(rational result)
			P_i = b_i*Q_i - P_im1;
			Q_ip1 = Q_im1 + b_i*(P_im1-P_i);
		} while (P_i != P_im1);
		
		// result
		BigInteger gcd = N.gcd(BigInteger.valueOf(P_i));
		return (gcd.compareTo(ONE)>0 && gcd.compareTo(N)<0) ? gcd : null;
	}
}
