// Written by Tilman Neumann (www.tilman-neumann.de) 2016, 2017
// Any commercial use of this software requires the written consent of the author.
// Usage and republication are free for non-commercial purposes, as long as this header is carried along.
// The author takes no warranties for the correctness of the software.
package de.tilman_neumann.types;

import java.util.Iterator;
import java.util.Map;

/**
 * A multiset with a sort relation between elements.<br>
 * 
 * The sorting of elements has the following consequences:
 * - elements must be Comparable
 * - SortedMaps are a bit slower than HashMaps
 * - output of sorted multisets looks nicer than that of unsorted multisets
 * 
 * Since elements are sorted, we can easily define a sorting on SortedMultisets, too:
 * A sorted multiset is bigger than another one if it's biggest element is bigger
 * than the largest element of the other multiset; or the 2.nd-biggest if
 * the biggest elements are equal; or the 3.rd biggest, and so on.
 * 
 * @author Tilman Neumann
 * 
 * @param <T> value class
 */
public interface SortedMultiset<T extends Comparable<T>> extends Multiset<T>, Comparable<SortedMultiset<T>> {

	/**
	 * @return The smallest element.
	 */
	T getSmallestElement();

	/**
	 * @return The biggest element.
	 */
	T getBiggestElement();
	
	/**
	 * @return an iterator that returns biggest elements first.
	 */
	Iterator<Map.Entry<T, Integer>> getTopDownIterator();
	
	/**
	 * Returns the multiset of elements contained in both this and in the other multiset.
	 * @param other
	 * @return
	 */
	@Override // declare exact result data type
	SortedMultiset<T> intersect(Multiset<T> other);
}